--[[ LibAccounts
  hooks for identifying which account you're on
  LibAccountsGlobal = {
    next_id = N,
    accounts = {
      acctid = nickname -- optional
    }
    shards = {
      shard = {
        chars = { char = acctid, ... },
        accounts = { acctid = { char = alliance, ... }, ... }
      }
    }
  }
  LibAccountsAccount = {
    id = N
  }
]]--

if not Library then Library = {} end
local info, Accounts = ...
Accounts.shard = "Unknown"
if not Library.LibAccounts then Library.LibAccounts = Accounts end

local slower = string.lower

Accounts.Version = "0.8-140216-18:22:39"
local printf = Library.printf.printf
local printfhs = Library.printf.printfhs
local sprintf = Library.printf.sprintf

Accounts.debugging = true

function Accounts.debug(...)
  if Accounts.debugging then
    printf(...)
  end
end

function Accounts.find_myself()
  -- Okay, let's see whether we can figure out who we are.
  local me = Inspect.Unit.Detail('player')
  if me and me.name and not Accounts.whoami then
    if me.name:find(' ') then
      -- because of this one Eastern Holdings quest.
      return
    end
    Accounts.whoami = slower(me.name)
    Accounts.whoami_alliance = me.alliance
    local acct = Accounts.here.chars[Accounts.whoami]
    Accounts.acctid = acct or LibAccountsAccount.id
  else
    Accounts.acctid = Accounts.acctid or LibAccountsAccount.id
  end
end

function Accounts.table_init()
  local ishard = Inspect.Shard()
  local shard = (ishard and ishard.name) or "Unknown"
  if shard ~= "Unknown" then
    Accounts.shard = shard
  else
    printf("Shard is unknown (ishard %s)", tostring(ishard))
  end
  if not LibAccountsGlobal.shards then
    LibAccountsGlobal.shards = {}
  end
  if not LibAccountsGlobal.accounts then
    LibAccountsGlobal.accounts = {}
  end
  if not LibAccountsGlobal.next_id then
    LibAccountsGlobal.next_id = 1
  end
  -- make sure next_id is actually a new account
  for k, v in pairs(LibAccountsGlobal.accounts) do
    if k >= LibAccountsGlobal.next_id then
      LibAccountsGlobal.next_id = k + 1
    end
  end
  if not LibAccountsGlobal.shards[shard] then
    LibAccountsGlobal.shards[shard] = {
      chars = {},
      accounts = {},
    }
  end
  Accounts.here = LibAccountsGlobal.shards[shard]
  -- try to figure out who we are...
  local found_entry = Accounts.find_myself()
  -- can't proceed if we don't have an account ID yet.
  if not Accounts.acctid then
    Accounts.debug("table_init: No acctid.")
    return
  end
  if not Accounts.here.accounts[Accounts.acctid] then
    Accounts.here.accounts[Accounts.acctid] = {}
  end
  if Accounts.whoami then
    if Accounts.here.chars[Accounts.whoami] ~= Accounts.acctid then
      Accounts.debug("%s: changing from %s to %s?",
      	Accounts.whoami,
	Accounts.acctname(Accounts.here.chars[Accounts.whoami]),
	Accounts.acctname(Accounts.acctid))
    end
    Accounts.here.chars[Accounts.whoami] = Accounts.acctid
    if Accounts.here.accounts[Accounts.acctid][Accounts.whoami] then
      if Accounts.here.accounts[Accounts.acctid][Accounts.whoami] ~= Accounts.whoami_alliance then
        printf("Apparently %s changed from %s to %s?",
	  Accounts.whoami,
          Accounts.here.accounts[Accounts.acctid][Accounts.whoami], Accounts.whoami_alliance)
      end
    end
    Accounts.here.accounts[Accounts.acctid][Accounts.whoami] = Accounts.whoami_alliance
  end
end

function Accounts.acctname(id)
  id = id or Accounts.acctid
  return LibAccountsGlobal.accounts[id] or sprintf("#%d", id or 0)
end

function Accounts.variables_loaded(event, name)
  if name ~= 'LibAccounts' then
    return
  end
  if not LibAccountsGlobal then
    LibAccountsGlobal = {
      next_id = 1,
      chars = {},
      accounts = {}
    }
  end
  if not LibAccountsAccount then
    LibAccountsAccount = {
      id = LibAccountsGlobal.next_id
    }
    LibAccountsGlobal.next_id = LibAccountsGlobal.next_id + 1
  end
  Accounts.table_init()
end

-- identifies all the characters on this shard that can be reached
-- from the current character
function Accounts.available_chars(acct_only)
  local availables = {}
  local visited = {}
  Accounts.table_init()
  -- First, we add everything on this account
  if not Accounts.here or not Accounts.here.accounts[Accounts.acctid] then
    printf("Warning:  Tried to find available chars, didn't find any.")
    return availables
  end
  Accounts.populate(availables, visited, Accounts.acctid, acct_only)
  return availables
end

function Accounts.available_p(charname, acct_only)
  local availables = Accounts.available_chars(acct_only)
  charname = slower(charname)
  if availables[charname] then
    return true
  end
  return false
end

function Accounts.populate(availables, visited, acctid, acct_only)
  local alliances = {}
  if visited[acctid] then
    return
  end
  if not Accounts.here.accounts or not Accounts.here.accounts[acctid] then
    return
  end
  visited[acctid] = true
  for char, alliance in pairs(Accounts.here.accounts[acctid]) do
    alliances[alliance] = true
    table.insert(availables, { char = char, alliance = alliance, acctid = acctid })
    availables[slower(char)] = { alliance = alliance, acctid = acctid }
  end
  if not acct_only then
    for other_acct, chars in pairs(Accounts.here.accounts) do
      local matched_alliance = false
      for char, alliance in pairs(chars) do
        if alliances[alliance] then
          matched_alliance = true
	  break
        end
      end
      if matched_alliance then
        Accounts.populate(availables, visited, other_acct, acct_only)
      end
    end
  end
end

function Accounts.acct_of(charname)
  if not charname then
    local me = Inspect.Unit.Detail("player")
    charname = (me and me.name) or "Unknown"
  end
  charname = slower(charname)
  Accounts.table_init()
  for acct, tab in pairs(Accounts.here.accounts) do
    for char, alliance in pairs(tab) do
      if slower(char) == charname then
        return acct
      end
    end
  end
  return nil
end

function Accounts.slashcommand(args)
  -- invalid arguments already diagnosed
  if not args then
    return
  end
  local did_something = false
  Accounts.table_init()
  if not Accounts.here then
    printf("... No info?")
    return
  end
  if args.f then
    -- force to integer
    args.f = math.floor(args.f)
    did_something = true
    if not Accounts.here.accounts[args.f] then
      if args.f > LibAccountsGlobal.next_id then
        printf("Account id %d too high, reducing to %d.",
	  args.f, LibAccountsGlobal.next_id)
        args.f = LibAccountsGlobal.next_id
      end
      if args.f == LibAccountsGlobal.next_id then
        LibAccountsGlobal.next_id = LibAccountsGlobal.next_id + 1
      end
      Accounts.here.accounts[args.f] = {}
    end
    if Accounts.acctid ~= args.f then
      Accounts.here.accounts[args.f][Accounts.whoami] = Accounts.here.accounts[Accounts.acctid][Accounts.whoami]
      Accounts.here.accounts[Accounts.acctid][Accounts.whoami] = nil
      printf("Moving '%s' from account '%s' to account '%s'.",
        Accounts.whoami, Accounts.acctname(), Accounts.acctname(args.f))
      Accounts.acctid = args.f
    else
      printf("Already on account %d.", args.f)
    end
    -- and change .chars too.
    if Accounts.here.chars[Accounts.whoami] ~= Accounts.acctid then
      Accounts.here.chars[Accounts.whoami] = Accounts.acctid
      printf("Flagging '%s' as part of account '%s'.",
        Accounts.whoami, Accounts.acctname())
    end
  end
  if args.n then
    did_something = true
    printf("Setting nickname for account '%s' to '%s'.",
      Accounts.acctname(), args.n)
    LibAccountsGlobal.accounts[Accounts.acctid] = args.n
  end
  if args.l then
    did_something = true
    Accounts.dump_acct(args.l)
  end
  if not did_something then
    printf("Dumping account info for %s:", Accounts.shard)
    if #args.leftover_args > 0 then
      for i = 1, #args.leftover_args do
	local name = slower(args.leftover_args[i])
	if Accounts.here.chars[name] then
	  printf("  %s: %s", name, Accounts.acctname(Accounts.here.chars[name]))
	else
          printf("  %s: no data", name)
	end
      end
    else
      for acct, tab in pairs(Accounts.here.accounts) do
	Accounts.dump_acct(acct)
      end
      printf("Available characters:")
      local chars = Accounts.available_chars()
      for _, row in ipairs(chars) do
	printf("  %s (%s, acct %s)", row.char, row.alliance, Accounts.acctname(row.acctid))
      end
    end
  end
end

function Accounts.dump_acct(id)
  local acct = math.floor(id)
  local tab = Accounts.here.accounts[acct]
  printf("Account %s:", Accounts.acctname(acct))
  if next(tab) then
    for char, alliance in pairs(tab) do
      printf("  %s (%s)", char, alliance)
    end
  else
    printf("No data for this account?")
  end
end

function Accounts.availability(event, chars)
  for k, v in pairs(chars) do
    if v == 'player' then
      Accounts.table_init()
      return
    end
  end
end

-- LibGetOpt is optional now.
if Library and Library.LibGetOpt then
  Library.LibGetOpt.makeslash("f:l:n:", "LibAccounts", "accounts", Accounts.slashcommand)
end

Command.Event.Attach(Event.Unit.Availability.Full, Accounts.availability, "full availability hook")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, Accounts.variables_loaded, "variable loaded hook")
